/****************************************************************************
*   serdump.c -- monitor a serial device as hex dump                        *
*   Copyright (C) 1998, 1999  Peter Mack, DL3FCJ                            *
*                                                                           *
*   This program is free software; you can redistribute it and/or modify    *
*   it under the terms of the GNU General Public License as published by    *
*   the Free Software Foundation; either version 2 of the License, or       *
*   (at your option) any later version.                                     *
*                                                                           *
*   This program is distributed in the hope that it will be useful,         *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
*   GNU General Public License for more details.                            *
*                                                                           *
*   You should have received a copy of the GNU General Public License       *
*   along with this program; if not, write to the                           *
*   Free Software Foundation, Inc.                                          *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.                *
*                                                                           *
****************************************************************************/

#include <sys/time.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include "lock.h"

#define BUFSIZE 1024


/*
 * global variables
 */
char	*version = "0.4";
static struct termios pots;	/* old port termios settings      */
static struct termios sots;	/* old stdout/in termios settings */
char *device;			/* device name                    */
int handle;			/* device handle                  */
int logfile;			/* handle of log file             */


/*
 * usage
 */
void usage(void)
{
	fprintf(stderr, "\nUsage: serdump <device> <speed> [<log file>]\n");
	fprintf(stderr,   "  e.g. serdump /dev/ttyS1 9600\n");
	fprintf(stderr,   "   or  serdump /dev/ttyS1 9600 test.log\n\n");
	exit(1);
}


/*
 * convert speed
 */
speed_t symbolic_speed(int speednum)
{ 
	if (speednum >= 460800)	return B460800;
	if (speednum >= 230400)	return B230400;
	if (speednum >= 115200)	return B115200;
	if (speednum >= 57600)	return B57600;
	if (speednum >= 38400)	return B38400;
	if (speednum >= 19200)	return B19200;
	if (speednum >= 9600)	return B9600;
	if (speednum >= 4800)	return B4800;
	if (speednum >= 2400)	return B2400;
	if (speednum >= 1800)	return B1800;
	if (speednum >= 1200)	return B1200;
	if (speednum >= 600)	return B600;
	if (speednum >= 300)	return B300;
	if (speednum >= 200)	return B200;
	if (speednum >= 150)	return B150;
	if (speednum >= 134)	return B134;
	if (speednum >= 110)	return B110;
	if (speednum >= 75)	return B75;
	
	return B50;
}


/*
 * restore original terminal settings on exit
 */
void cleanup_termios(int signal)
{
	if (logfile > 0) {
		close(logfile);
		logfile = 0;
	}
	
	close(handle);
	unlock_device(device);

	tcsetattr(handle, TCSANOW, &pots);
	tcsetattr(STDIN_FILENO, TCSANOW, &sots);

	printf("\n\n");
   
	exit(0);
} 


/*
 * assemble a line and print it
 */
void printline(char *rbuf, int rnum, int chars)
{
char buf[85];
char xbuf[5];
int pos;
int i;

	sprintf(buf, "\r%08X:  ", ((chars - 1) / 16) * 16);
	pos = 12;
	
	for (i = 0; i <= rnum - 1; i++) {
		sprintf(xbuf, "%02X ", rbuf[i]);
		strcat(buf, xbuf);
		pos += 3;
		if (!((i + 1) % 8)) {
			strcat(buf, " ");
			pos++;
		}
	}


	for (i = 0; i < (63 - pos); i++) {
		strcat(buf, " ");
	}

	for (i = 0; i <= rnum - 1; i++) {
		if ((rbuf[i] < 32) || (rbuf[i] > 128)) {
			strcat(buf, ".");
		}
		else {
			sprintf(xbuf, "%c", rbuf[i]);
			strcat(buf, xbuf);
		}
		if (!((i + 1) % 8)) {
			strcat(buf, " ");
		}
	}


	if (rnum >= 16) {
		strcat(buf, "\n");
		if (logfile > 0) {
			write(logfile, buf, strlen(buf));
		}
	}
	
	write(STDOUT_FILENO, buf, strlen(buf));

}


/*
 * main
 */
int main(int argc, char *argv[])
{
struct termios pts;	/* termios settings on port              */
struct termios sts;	/* termios settings on stdout/in         */
int speed;		/* used in argument parsing to set speed */
struct sigaction sact;	/* used to initialize the signal handler */
fd_set ready;		/* used for select                       */
int i;			/* used in the multiplex loop            */
int j;			/* loop variable                         */
int done;
char buf[BUFSIZE];	/* receive buffer                        */
long int chars;		/* total number of chars received        */
char pbuf[16];		/* print buffer                          */
int pchars;		/* number of chars in the printbuffer    */

	logfile = 0;	/* for sure! */

	printf("This is SerDump version %s.\n"
	       "Monitor a serial device as hex dump.\n"
	       "(c) copyright 1998, 1999  Peter Mack, DL3FCJ\n", version);

	if (argc < 3)
		usage();

	printf("\n");

	device = argv[1];
	
	if (lock_device(device) < 0) {
		printf("Device %s locked.\n", device);
		return -1;
	}

	if ((handle = open(device, O_RDWR)) == -1) {
		fprintf(stderr, "ERROR: can't open serial port\n");
		unlock_device(device);
		return -1;
	}
	
	printf("Press any key to end program!\n\n");
	
	if (argc == 4) {
		logfile = open(argv[3], O_RDWR | O_CREAT, 0644);
		if (logfile < 0) {
			printf("Sorry, can not open log file: %s\n", argv[3]);
		}
	}

	speed = atoi(argv[2]);

	/* modify the port configuration */
	tcgetattr(handle, &pts);
	pots = pts;
	/* some things we want to set arbitrarily */
	pts.c_lflag &= ~ICANON; 
	pts.c_lflag &= ~(ECHO | ECHOCTL | ECHONL);
	pts.c_cflag |= HUPCL;
	pts.c_cc[VMIN] = 1;
	pts.c_cc[VTIME] = 0;

	/* Standard CR/LF handling: this is a dumb terminal.
	 * Do no translation:
	 *  no NL -> CR/NL mapping on output, and
	 *  no CR -> NL mapping on input.
	 */
	pts.c_oflag &= ~ONLCR;
	pts.c_iflag &= ~ICRNL;

	cfsetospeed(&pts, symbolic_speed(speed));
	cfsetispeed(&pts, symbolic_speed(speed));


	/* Now deal with the local terminal side */
	tcgetattr(STDIN_FILENO, &sts);
	sots = sts;
	/* again, some arbitrary things */
	sts.c_iflag &= ~BRKINT;
	sts.c_iflag |= IGNBRK;
	sts.c_lflag &= ~ISIG;
	sts.c_cc[VMIN] = 1;
	sts.c_cc[VTIME] = 0;
	sts.c_lflag &= ~ICANON;
	/* no local echo: allow the other end to do the echoing */
	sts.c_lflag &= ~(ECHO | ECHOCTL | ECHONL);


	/*
	 * set the signal handler to restore the old
	 * termios handler
	 */
	sact.sa_handler = cleanup_termios; 
	sigaction(SIGHUP, &sact, NULL);
	sigaction(SIGINT, &sact, NULL);
	sigaction(SIGPIPE, &sact, NULL);
	sigaction(SIGTERM, &sact, NULL);

	/* Now set the modified termios settings */
	tcsetattr(handle, TCSANOW, &pts);
	tcsetattr(STDIN_FILENO, TCSANOW, &sts);

	done = 0;
	chars = 0;
	pchars = 0;
   
	do {
		FD_ZERO(&ready);
		FD_SET(STDIN_FILENO, &ready);
		FD_SET(handle, &ready);

		if (select(FD_SETSIZE, &ready, NULL, NULL, NULL) < 0) {
			perror("select");
			return 1;
		}
      
		if (FD_ISSET(handle, &ready)) {
			/* handle has characters for us */
			i = read(handle, buf, BUFSIZE);
			if (i >= 1) {
				for (j = 0; j < i; j++ ) {
					pbuf[pchars++] = buf[j];
					chars++;
					if (pchars >= 16) {
						printline(pbuf, pchars, chars);
						pchars = 0;
					}
				}
				if (pchars)
					printline(pbuf, pchars, chars);
			} else {
				done = 1;
			}
		}
      
		if (FD_ISSET(STDIN_FILENO, &ready)) {
			/* standard input has characters for us */
			i = read(STDIN_FILENO, buf, BUFSIZE);
			if (i >= 1) {
				done = 1;
			}
		}
	} while (!done);
	
	if (logfile > 0) {
		close(logfile);
		logfile = 0;
	}
	
	close(handle);
	unlock_device(device);

	/* restore original terminal settings and exit */
	tcsetattr(handle, TCSANOW, &pots);
	tcsetattr(STDIN_FILENO, TCSANOW, &sots);

	printf("\n\n");
   
	exit(0);
}
